@@ -4,8 +4,6 @@
 ** This program is in the public domain and may be used freely by anyone
 ** who wants to.
 **
-** Last update: 11 Aug 1996
-**
 ** Please send bug fixes/bug reports to: Peter Eriksson <pen@lysator.liu.se>
 */
 
@@ -79,8 +77,9 @@ extern void *malloc();
 struct nlist nl[] =
 {
 #define N_TCB 0
-
   { "_tcb" },
+#define N_BTEXT 1
+  { "_btext" },
   { "" }
 };
 
@@ -117,21 +116,32 @@ int k_open()
 ** Returns 1 if call succeeded, else 0 (zero).
 */
 static int getbuf(addr, buf, len, what)
-  long addr;
+  unsigned long addr;
   char *buf;
-  int len;
+  unsigned int len;
   char *what;
 {
-  if (kvm_read(kd, addr, buf, len) < 0)
-  {
-    if (syslog_flag)
-	syslog(LOG_ERR, "getbuf: kvm_read(%08x, %d) - %s : %m",
-	     addr, len, what);
-
-    return 0;
-  }
-
-  return 1;
+    if (addr < nl[N_BTEXT].n_value ||		/* Overkill.. */
+	addr >= (unsigned long)0xFFC00000 ||
+	(addr + len) < nl[N_BTEXT].n_value ||
+	(addr + len) >= (unsigned long)0xFFC00000)
+    {
+	if (syslog_flag)
+	    syslog(LOG_ERR, "getbuf: bad address (%08x not in %08x-0xFFC00000) - %s",
+		   addr, nl[N_BTEXT].n_value, what);
+	return 0;
+    }
+    
+    if (kvm_read(kd, addr, buf, len) < 0)
+    {
+	if (syslog_flag)
+	    syslog(LOG_ERR, "getbuf: kvm_read(%08x, %d) - %s : %m",
+		   addr, len, what);
+	
+	return 0;
+    }
+    
+    return 1;
 }
 
 /*
@@ -170,9 +180,9 @@ static struct socket *
 #ifdef _HAVE_OLD_INPCB
   head = pcbp->inp_prev;
 #endif
+#ifdef _HAVE_OLD_INPCB
   do
   {
-#ifdef _HAVE_OLD_INPCB
     if ( pcbp->inp_faddr.s_addr == faddr->s_addr &&
 	 pcbp->inp_laddr.s_addr == laddr->s_addr &&
 	 pcbp->inp_fport        == fport &&
@@ -184,6 +194,8 @@ static struct socket *
 		  sizeof(struct inpcb),
 		  "tcblist"));
 #else
+  do
+  {
     if (!getbuf((long) head, &pcbp, sizeof(struct inpcb), "tcblist"))
        break;
     if (pcbp.inp_faddr.s_addr == faddr->s_addr &&
@@ -194,7 +206,7 @@ static struct socket *
   } while (head != NULL);
 #endif
 
-  return NULL;
+return NULL;
 }
 
 /*
@@ -248,7 +260,11 @@ int k_getuid(faddr, fport, laddr, lport, uid)
         return(-1);
 
       ofiles = (struct file **) malloc(pfd.fd_nfiles * sizeof(struct file *));
-
+      if (!ofiles) {
+	  ERROR("k_getuid: ofiles malloc failed");
+	  return(-1);
+      }
+      
       if(!getbuf(pfd.fd_ofiles, ofiles,
                  pfd.fd_nfiles * sizeof(struct file *), "ofiles"))
       {
@@ -258,11 +274,14 @@ int k_getuid(faddr, fport, laddr, lport, uid)
 
       for(j = 0; j < pfd.fd_nfiles; j ++)
       {
-        if(!getbuf(ofiles[j], &ofile, sizeof(struct file), "ofile"))
-	{
-          free(ofiles);
-          return(-1);
-        }
+	if (!ofiles[j])	/* might be sparse */
+	  continue;
+	  
+	  if(!getbuf(ofiles[j], &ofile, sizeof(struct file), "ofile"))
+	  {
+	      free(ofiles);
+	      return(-1);
+	  }
 
         if(ofile.f_count == 0)
 	  continue;
