@@ -6,8 +6,6 @@
 **
 ** Author: Casper Dik <casper@fwi.uva.nl>
 **
-** Last update: 13 Oct 1994
-**
 ** Please send bug fixes/bug reports to: Peter Eriksson <pen@lysator.liu.se>
 */
 
@@ -42,6 +40,7 @@
 
 #ifdef ipc_tcp_laddr
 #define SOLARIS24 1
+#define BROKEN_HASH
 /*
  * In Solaris 2.4 there have been a number of changes:
  * - the ipc_s structure had its field names changed
@@ -66,23 +65,11 @@
 #ifndef SOLARIS24
 #define N_FILE	 1
 #endif
-#ifdef SOLARIS24
-#define NEED_OWN_KVMPROC /* bug workaround, kvm_setproc doesn't show
-			  * processes started after kvmopen */
-#endif
 
 struct nlist nl[] = {
 	{ "ipc_tcp_fanout" },
 #ifndef SOLARIS24
 	{ "file" },
-#endif
-#ifdef NEED_OWN_KVMPROC
-	{ "practive" },
-#ifdef N_FILE
-#define N_PRACTIVE 2
-#else
-#define N_PRACTIVE 1
-#endif
 #endif
 	{ 0 },
 };
@@ -92,6 +79,13 @@ static kvm_t *kd;
 
 
 int k_open()
+#ifdef SOLARIS24_WORKAROUND
+{
+    return 0;
+}
+
+static int local_k_open()
+#endif
 {
   /*
   ** Open the kernel memory device
@@ -104,10 +98,16 @@ int k_open()
   */
   if (kvm_nlist(kd, nl) != 0)
     ERROR("main: kvm_nlist");
-
+  
   return 0;
 }
 
+#ifdef SOLARIS24_WORKAROUND
+static void local_k_close()
+{
+    kvm_close(kd);
+}
+#endif
 
 /*
 ** Get a piece of kernel memory with error handling.
@@ -119,17 +119,17 @@ static int getbuf(addr, buf, len, what)
   int len;
   char *what;
 {
+    int i, status;
 
-  if (kvm_read(kd, addr, buf, len) < 0)
-  {
-    if (syslog_flag)
-      syslog(LOG_ERR, "getbuf: kvm_read(%08x, %d) - %s : %m",
-	     addr, len, what);
 
-    return 0;
-  }
+    i = 0;
+    while (i < 10 && (status = kvm_read(kd, addr, buf, len)) < 0)
+	++i;
+
+    if (status < 0)
+	return 0;
   
-  return 1;
+    return 1;
 }
 
 
@@ -142,6 +142,27 @@ int k_getuid(faddr, fport, laddr, lport, uid)
   struct in_addr *laddr;
   int lport;
   int *uid;
+#ifdef SOLARIS24_WORKAROUND
+{
+    extern int local_k_getuid();
+    int result;
+
+    
+    local_k_open();
+    result = local_k_getuid(faddr, fport, laddr, lport, uid);
+    local_k_close();
+
+    return result;
+}
+
+
+static int local_k_getuid(faddr, fport, laddr, lport, uid)
+  struct in_addr *faddr;
+  int fport;
+  struct in_addr *laddr;
+  int lport;
+  int *uid;
+#endif
 {
     queue_t sqr
 #ifndef SOLARIS24
@@ -164,11 +185,16 @@ int k_getuid(faddr, fport, laddr, lport, uid)
 #ifdef SOLARIS24
     struct proc *procp;
 #endif
+#ifdef BROKEN_HASH
+    ipc_t *alticp = 0;
+    unsigned int altoffset;
+#endif
     
     usfp = fport;
     uslp = lport;
 
-#ifdef SOLARIS24
+#ifdef BROKEN_HASH
+    /* code used (ports > 8) instead of (ports >> 8)
     /* low byte of local port number not used, low byte of 
        local addres is used
 	ip_bind  in the kernel (+ approx 0x4c0)
@@ -184,18 +210,23 @@ int k_getuid(faddr, fport, laddr, lport, uid)
 
      */
 #if (defined(BIG_ENDIAN) || defined(_BIG_ENDIAN))
-    offset = usfp >> 8;
+    altoffset = usfp >> 8;
 #else
-    offset = uslp >> 8;
+    altoffset = uslp >> 8;
 #endif
-    offset ^= usfp ^ uslp;
-    offset ^= faddr->S_un.S_un_b.s_b4;
+    altoffset ^= usfp ^ uslp;
+    altoffset ^= faddr->S_un.S_un_b.s_b4;
     if (uslp > 8 || usfp != 0)
-	offset ^= 1;
-#else
+	altoffset ^= 1;
+    altoffset &= 0xff;
+    if (!getbuf(nl[N_FANOUT].n_value + sizeof(ipc_t *) * altoffset,
+		(char *) &alticp,
+		sizeof(ipc_t *),
+		"ipc_tcp_fanout[altoffset]"))
+	alticp = 0;
+#endif
     offset = usfp ^ uslp;
     offset ^= (unsigned) faddr->S_un.S_un_b.s_b4 ^ (offset >> 8);
-#endif
     offset &= 0xff;
 
     if (!getbuf(nl[N_FANOUT].n_value + sizeof(ipc_t *) * offset,
@@ -204,6 +235,12 @@ int k_getuid(faddr, fport, laddr, lport, uid)
 		"ipc_tcp_fanout[offset]"))
 	return -1;
     
+#ifdef BROKEN_HASH
+    if (icp == 0 && alticp != 0) {
+	icp = alticp;
+	alticp = 0;
+    }
+#endif
 #ifndef DEBUGHASH
     if (icp == 0) {
 	syslog(LOG_INFO, "k_getuid: Hash miss");
@@ -255,6 +292,12 @@ int k_getuid(faddr, fport, laddr, lport, uid)
 	    memcmp(&faddr->s_addr, raddr, 4) == 0)
 		break;
 	icp = ic.ipc_hash_next;
+#ifdef BROKEN_HASH
+	if (icp == 0 && alticp != 0) {
+	    icp = alticp;
+	    alticp = 0;
+	}
+#endif
     }
 #ifdef DEBUGHASH
     if (icp)
@@ -289,7 +332,7 @@ int k_getuid(faddr, fport, laddr, lport, uid)
     if (kvm_setproc(kd) != 0)
 	return -1;
 
-    while (procp = kvm_nextproc(kd)) {
+    while ((procp = kvm_nextproc(kd)) != NULL) {
 	struct uf_entry files[NFPCHUNK];
 	int nfiles = procp->p_user.u_nofiles;
 	unsigned long addr = (unsigned long) procp->p_user.u_flist;
@@ -367,31 +410,3 @@ int k_getuid(faddr, fport, laddr, lport, uid)
 #endif
     return -1;
 }
-
-#ifdef NEED_OWN_KVMPROC
-/* Work around for Solaris 2.x bug in kvm_setproc,
- * kvm_setproc doesn't reread practive */
-
-static struct proc *nextp, currentp;
-
-int kvm_setproc (kvm_t *kd)
-{
-    if (!getbuf((unsigned long) nl[N_PRACTIVE].n_value,
-		(char *) &nextp,
-		sizeof(nextp),
-		"practive"))
-	return -1;
-    return 0;
-}
-
-struct proc *kvm_nextproc(kvm_t *kd)
-{
-    if (!getbuf((unsigned long) nextp,
-		(char*) &currentp,
-		sizeof(currentp),
-		"kvm_getproc"))
-	return 0;
-    nextp = currentp.p_next;
-    return &currentp;
-}
-#endif
