@@ -154,6 +154,25 @@ static int check_noident(homedir)
   return (rcode == 0);
 }
 
+#ifdef INCLUDE_CRYPT
+/*
+** Checks address of incoming call against network/mask pairs of trusted
+** networks to determine whether to crypt response or not.
+*/
+int check_crypt(faddr)
+  struct in_addr *faddr;
+{
+  int i;
+  extern int netcnt;
+  extern u_long localnet[], localmask[];
+ 
+  for (i = 0; i < netcnt; i++) {
+    if ((faddr->s_addr & localmask[i]) == localnet[i])
+      return 0;
+  }
+  return 1;
+}
+#endif
 
 int parse(fp, laddr, faddr)
   FILE *fp;
@@ -552,11 +571,12 @@ int parse(fp, laddr, faddr)
 
 #ifdef ALLOW_FORMAT
     grp = getgrgid(pwp->pw_gid);
-    if (!grp || pwp->pw_gid != grp->gr_gid)
+    if (grp && pwp->pw_gid != grp->gr_gid)
     {
-      if (syslog_flag)
-	syslog(LOG_WARNING, "getgrgid() could not map gid (%d) to name (for uid %d, name %s)",
-	       pwp->pw_gid, uid, pwp->pw_name);
+	if (syslog_flag)
+	    syslog(LOG_WARNING,
+		   "getgrgid() could not map gid (%d) to name (for uid %d, name %s)",
+		   pwp->pw_gid, uid, pwp->pw_name);
 
       printf("%d , %d : USERID : OTHER%s%s : %d\r\n",
 	     lport, fport,
@@ -565,7 +585,10 @@ int parse(fp, laddr, faddr)
 	     uid);
       continue;
     }
-    sprintf (grname, "%.99s", grp->gr_name);
+    if (grp)
+	sprintf (grname, "%.99s", grp->gr_name);
+    else
+	sprintf (grname, "%d", pwp->pw_gid);
 #endif
 
     /*
@@ -595,7 +618,7 @@ int parse(fp, laddr, faddr)
     }
 
 #ifdef INCLUDE_CRYPT
-    if (crypto_flag)
+    if (crypto_flag && check_crypt(faddr))
       printf("%d , %d : USERID : OTHER%s%s : [%s]\r\n",
 	     lport, fport,
 	     charset_name ? " , " : "",
